from pydantic import BaseModel, Field
from redis import Redis
from time import time
from typing import Tuple, Optional
import json


class RedisRateLimiterArgs(BaseModel):
    host: str = Field(default="localhost", description="Redis server host")
    port: int = Field(default=6379, description="Redis server port")
    db: int = Field(default=0, description="Redis database index")
    window: int = Field(default=60, description="Time window in seconds")
    limit: int = Field(default=100, description="Maximum number of tokens per window (максимальный потолок токенов)")

    class Config:
        strict = True


class RedisRateLimiter:
    """
    Rate limiter на основе алгоритма Token Bucket с использованием Redis.
    Поддерживает динамическое изменение лимитов и временные правила.

    Каждый клиент имеет свой bucket с токенами, который пополняется со временем.
    Токены добавляются пропорционально прошедшему времени, но не могут превышать limit.

    Пример:
        limiter = RedisRateLimiter(RedisRateLimiterArgs(
            limit=100,    # максимум 100 токенов
            window=60     # восполнение за 60 секунд
        ))

        # Даже после долгого простоя, токенов не будет больше 100
        await limiter.acquire("user_123")  # Успех, осталось 99 токенов
        await limiter.acquire("user_123", tokens=50)  # Успех, осталось 49 токенов
        await limiter.acquire("user_123", tokens=50)  # Неудача, токенов недостаточно
    """

    def __init__(self, args: RedisRateLimiterArgs):
        self.args = args
        self.redis = Redis(host=args.host, port=args.port, db=args.db, decode_responses=True)
        self.refill_rate = args.limit / args.window

    def _get_bucket_key(self, client_id: str) -> str:
        """Формирует ключ для хранения состояния bucket клиента."""
        return f"rate_limit:{client_id}:bucket"

    def _get_rules_key(self, client_id: str) -> str:
        """Формирует ключ для хранения правил клиента."""
        return f"rate_limit:{client_id}:rules"

    def _get_bucket(self, client_id: str) -> tuple[float, float]:
        """
        Получает текущее состояние бакета.

        Args:
            client_id: ID клиента

        Returns:
            Кортеж (текущие токены, время последнего обновления)
        """
        key = self._get_bucket_key(client_id)
        data = self.redis.mget([f"{key}:tokens", f"{key}:last_update"])

        if not data[0]:  # Если bucket не существует
            return float(self.get_current_limit(client_id)), time()

        return float(data[0]), float(data[1])

    def _update_bucket(self, client_id: str, tokens: float, last_update: float) -> None:
        """
        Обновляет состояние бакета.

        Args:
            client_id: ID клиента
            tokens: Новое количество токенов
            last_update: Время последнего обновления
        """
        key = self._get_bucket_key(client_id)
        pipeline = self.redis.pipeline()
        pipeline.mset({
            f"{key}:tokens": str(tokens),
            f"{key}:last_update": str(last_update)
        })
        pipeline.expire(key, self.args.window * 2)
        pipeline.execute()

    def get_current_limit(self, client_id: str) -> int:
        """
        Получает текущий лимит для клиента с учетом всех правил.

        Args:
            client_id: ID клиента

        Returns:
            Текущий лимит токенов
        """
        rules_key = self._get_rules_key(client_id)
        current_time = time()

        # Получаем правила
        rules_data = self.redis.get(rules_key)
        if rules_data:
            try:
                rules = json.loads(rules_data)
                # Фильтруем истекшие правила
                active_rules = [rule for rule in rules if rule["expires_at"] > current_time]

                if active_rules:
                    # Возвращаем минимальный лимит из активных правил
                    return min(rule["limit"] for rule in active_rules)
            except (json.JSONDecodeError, KeyError, TypeError):
                pass

        # Если нет активных правил, возвращаем базовый лимит
        return self.args.limit

    def set_permanent_limit(self, client_id: str, new_limit: int) -> None:
        """
        Устанавливает новый постоянный лимит для клиента.

        Args:
            client_id: ID клиента
            new_limit: Новый лимит токенов
        """
        self.args.limit = new_limit
        self.refill_rate = new_limit / self.args.window

        # Обновляем текущее состояние
        current_tokens, last_update = self._get_bucket(client_id)
        new_tokens = min(current_tokens, new_limit)
        self._update_bucket(client_id, new_tokens, last_update)

    def add_temporary_limit(self, client_id: str, limit: int, duration: int) -> None:
        """
        Добавляет временное правило лимита.

        Args:
            client_id: ID клиента
            limit: Временный лимит токенов
            duration: Продолжительность в секундах
        """
        rules_key = self._get_rules_key(client_id)
        current_time = time()
        expires_at = current_time + duration

        # Получаем текущие правила
        rules_data = self.redis.get(rules_key)
        if rules_data:
            try:
                rules = json.loads(rules_data)
                # Очищаем истекшие правила
                rules = [rule for rule in rules if rule["expires_at"] > current_time]
            except (json.JSONDecodeError, KeyError, TypeError):
                rules = []
        else:
            rules = []

        # Добавляем новое правило
        rules.append({
            "limit": limit,
            "expires_at": expires_at
        })

        # Сортируем правила по времени истечения (по убыванию) и лимиту
        rules.sort(key=lambda x: (-x["expires_at"], x["limit"]))

        # Сохраняем правила
        self.redis.set(rules_key, json.dumps(rules))

        # Устанавливаем TTL равным максимальной длительности
        if rules:
            max_expires = rules[0]["expires_at"]
            ttl = int(max_expires - current_time)
            if ttl > 0:
                self.redis.expire(rules_key, ttl)

        # Обновляем текущее состояние если новый лимит строже текущего
        current_tokens, last_update = self._get_bucket(client_id)
        if current_tokens > limit:
            self._update_bucket(client_id, float(limit), last_update)

    def remove_all_limits(self, client_id: str) -> None:
        """
        Удаляет все временные правила для клиента.
        """
        rules_key = self._get_rules_key(client_id)
        self.redis.delete(rules_key)

    def acquire(self, client_id: str, tokens: int = 1) -> bool:
        """
        Пытается получить токены из bucket клиента.

        Args:
            client_id: Уникальный идентификатор клиента
            tokens: Количество требуемых токенов (например, 1 для обычного запроса, больше для сложных операций)

        Returns:
            bool: True если токены получены, False если превышен лимит

        Примечание:
            Максимальное количество токенов ограничено параметром limit.
            Даже после долгого простоя, клиент не сможет накопить больше токенов, чем указано в limit.
        """
        current_tokens, last_update = self._get_bucket(client_id)
        now = time()

        # Получаем актуальный лимит
        current_limit = self.get_current_limit(client_id)
        self.refill_rate = current_limit / self.args.window

        elapsed = now - last_update
        new_tokens = min(
            current_limit,  # Используем актуальный лимит
            current_tokens + (elapsed * self.refill_rate)
        )

        if new_tokens < tokens:
            self._update_bucket(client_id, new_tokens, now)
            return False

        remaining = new_tokens - tokens
        self._update_bucket(client_id, remaining, now)
        return True

    def get_remaining_tokens(self, client_id: str) -> float:
        """
        Возвращает оставшееся количество токенов у клиента.

        Примечание:
            Возвращаемое значение никогда не превысит limit, даже если клиент
            не делал запросы очень долгое время.
        """
        current_tokens, last_update = self._get_bucket(client_id)
        now = time()
        elapsed = now - last_update

        # Используем актуальный лимит
        current_limit = self.get_current_limit(client_id)
        self.refill_rate = current_limit / self.args.window

        return min(
            current_limit,
            current_tokens + (elapsed * self.refill_rate)
        )

    async def demo_limit(self, client_id: str) -> None:
        """
        Демонстрация работы лимита токенов.
        """
        print(f"Начальное количество токенов: {await self.get_remaining_tokens(client_id)}")

        # Пробуем взять много токенов
        success = await self.acquire(client_id, tokens=self.args.limit)
        print(f"Попытка взять {self.args.limit} токенов: {'успех' if success else 'неудача'}")
        print(f"Осталось токенов: {await self.get_remaining_tokens(client_id)}")

        # Пробуем взять ещё
        success = await self.acquire(client_id, tokens=1)
        print(f"Попытка взять ещё 1 токен: {'успех' if success else 'неудача'}")
        print(f"Осталось токенов: {await self.get_remaining_tokens(client_id)}")

    def clear_limits(self, client_id: str) -> None:
        """
        Очищает все правила лимитов для клиента.

        Args:
            client_id: ID клиента
        """
        rules_key = self._get_rules_key(client_id)
        bucket_key = self._get_bucket_key(client_id)

        # Удаляем правила и текущее состояние
        self.redis.delete(rules_key)
        self.redis.delete(bucket_key)
